/*
Copyright (c) 2006 by Juliusz Chroboczek

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
*/

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <net/if.h>
#include <netinet/in.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <sys/time.h>
#include <time.h>
#include <errno.h>
#include <signal.h>
#include <endian.h>
#include <byteswap.h>

#include "ahcp-protocol.h"

char *data_file="/var/cache/ahcp.dat";
char *config_script = "/usr/local/bin/ahcp-config.sh";
char *interface = "eth1";
char *network_name = NULL;
int nodns = 0, debuglevel = 1;
int persistent = 0;
volatile sig_atomic_t exiting = 0, reconfigure = 0;

#define REQUEST_DELAY 10
#define REQUESTED_DELAY 60
#define REPLY_DELAY 20
#define UNSOLICITED_DELAY 3600

#define STATE_REQUESTING 0
#define STATE_REQUESTED 1
#define STATE_REPLY 2
#define STATE_IDLE 3

void
die(const char *string)
{
    fprintf(stderr, "%s", string);
    exit(1);
}

void
edie(char *string)
{
    perror(string);
    exit(1);
}

static int
draw(int bound)
{
    return random() % bound;
}

static int
draw2(int bound)
{
    return bound / 2 + random() % (bound / 2);
}

static unsigned long long
ntohll(unsigned long long value)
{
#if BYTE_ORDER == LITTLE_ENDIAN
    return bswap_64(value);
#else
    return value;
#endif
}

static int
timeout(int state)
{
    if(state == STATE_REQUESTING)
        return draw2(REQUEST_DELAY);
    else if(state == STATE_REQUESTED)
        return draw(REQUESTED_DELAY);
    else if(state == STATE_REPLY)
        return draw(REPLY_DELAY);
    else if(state == STATE_IDLE)
        return draw(UNSOLICITED_DELAY);
    else
        abort();
}

static int
id_message(unsigned char *buf, int len,
           unsigned long long *serial_return,
           unsigned int *ttl_return)
{
    union {
        unsigned int i;
        char c[4];
    } fourbytes;
    union {
        unsigned long long i;
        char c[8];
    } eightbytes;

    if(len < 13) {
        fprintf(stderr, "Message of length %d\n", len);
        return -1;
    }

    if(buf[0] != PROTOCOL_VERSION)
        return -1;
    
    memcpy(eightbytes.c, buf + 1, 8);
    if(serial_return)
        *serial_return = ntohll(eightbytes.i);

    memcpy(fourbytes.c, buf + 9, 4);
    if(ttl_return)
        *ttl_return = ntohl(fourbytes.i);
    
    return 1;
}
    

static int
recv_message(int s, unsigned long long *serial_return,
             unsigned int *ttl_return,
             unsigned char *buf, int *len_return)
{
    int rc;
    int len;

    len = recv(s, buf, 2048, MSG_DONTWAIT);
    if(len < 0) {
        perror("recv");
        return -1;
    }

    rc = id_message(buf, len, serial_return, ttl_return);
    if(rc >= 0) {
        if(len_return)
            *len_return = len;
    }
    return rc;
}

static int
send_message(int s, struct sockaddr_in6 *to, unsigned char *buf, int len)
{
    int rc;
    rc = sendto(s, buf, len, MSG_DONTWAIT, (struct sockaddr*)to, sizeof(*to));
    if(rc < 0) perror("write");
    return (rc == len ? 1 : -1);
}

static int
waitfd(int fd, int timeout)
{
    struct timeval tv1, tv2;
    fd_set r, e;
    int waited, rc;

    FD_ZERO(&r);
    FD_SET(fd, &r);

    FD_ZERO(&e);
    FD_SET(fd, &e);

    if(timeout < 0)
        timeout = 0;
    tv2.tv_sec = timeout;
    tv2.tv_usec = 0;
    gettimeofday(&tv1, NULL);
    rc = select(fd + 1, &r, NULL, &e, &tv2);
    gettimeofday(&tv2, NULL);

    if(rc < 0 && errno != EINTR) {
        perror("select");
        sleep(1);
        return 0;
    } else if(rc == 0) {
        return -1;
    }

    waited = tv2.tv_sec - tv1.tv_sec;
    if(tv2.tv_usec - tv1.tv_usec > 500000)
        waited += 1;

    return waited > timeout ? 0 : timeout - waited;
}

static char *
ntoa6(const unsigned char *data)
{
    const char *p;
    char buf1[17], buf2[100];
    memcpy(buf1, data, 16);
    buf1[16] = '\0';
    p = inet_ntop(AF_INET6, buf1, buf2, 100);
    if(p == NULL) return NULL;
    return strdup(p);
}

static char *
parse_address_list(const unsigned char *data, int len)
{
    char *result = NULL, *value;
    int i;

    if(len % 16 != 0)
        return NULL;

    i = 0;
    while(i < len) {
        value = ntoa6(data + i);
        if(value == NULL) {
            fprintf(stderr, "Couldn't grok IPv6 address");
            if(result) free(result);
            return NULL;
        }
        if(result == NULL) {
            result = value;
        } else {
            char *old_result = result;
            result = malloc(strlen(old_result) + 1 + strlen(value) + 1);
            strcpy(result, old_result);
            strcat(result, " ");
            strcat(result, value);
            free(old_result);
        }
        i += 16;
    }
    return result;
}
        
static int
parse_message(const unsigned char *data, int len, int start)
{
    int i = 4, opt, olen;
    pid_t pid;
    char *prefix = NULL, *nameserver = NULL;
    int routing_protocol = 0;
    char *routing_protocol_name = NULL;
    char *static_default_gw = NULL;
    char *olsr_multicast_address = NULL;
    unsigned long long serial;
    unsigned int ttl;

    union {
        unsigned int i;
        char c[4];
    } fourbytes;
    union {
        unsigned long long i;
        char c[8];
    } eightbytes;

    if(len < 13)
        return -2;

    if(data[0] != PROTOCOL_VERSION)
        return -1;
    
    memcpy(eightbytes.c, data + 1, 8);
    serial = ntohll(eightbytes.i);

    memcpy(fourbytes.c, data + 9, 4);
    ttl = ntohl(fourbytes.i);
    
    i = 13;

    while(1) {
        if(i >= len - 2) {
            fprintf(stderr, "Truncated reply\n");
            return -2;
        }
        opt = data[i];
        olen = data[i + 1];
        if(olen < 2 || olen + i > len) {
            fprintf(stderr, "Truncated reply\n");
            return -2;
        }
        if(opt == OPT_NETWORK_NAME) {
            if(network_name) {
                if(olen != strlen(network_name) + 2) {
                    return -1;
                }
                if(memcmp(data + i + 2, network_name, olen - 2) != 0)
                    return -1;
            }
        } else if(opt == OPT_SIGNATURE) {
            unsigned int slen;
            if(olen < 5)
                fprintf(stderr, "Warning: unexpected length for SIGNATURE.\n");
            if(data[i + 2] > SIGNATURE_RSA)
                fprintf(stderr, "Warning: unexpected signature type %d.\n",
                        data[i + 2]);
            slen = data[i + 3] << 8 | data[i + 4];
            if(slen > i + 5)
                fprintf(stderr, "Warning: unexpected signature length %d.\n",
                        slen);
            break;
        } else if(opt == OPT_IPv6_PREFIX || opt == OPT_NAME_SERVER) {
            char *value;
            if(olen % 16 != 2) {
                fprintf(stderr, "Unexpected length for %s.\n",
                        opt == OPT_IPv6_PREFIX ? "prefix" : "name server");
                return -2;
            }
            value = parse_address_list(data + i + 2, olen - 2);
            if(opt == OPT_IPv6_PREFIX)
                prefix = value;
            else
                nameserver = value;
        } else if(opt == OPT_ROUTING_PROTOCOL) {
            int j;
            if(olen < 3) {
                fprintf(stderr, "Unexpected size for routing protocol.\n");
                return -2;
            }
            routing_protocol = data[i + 2];
            if(routing_protocol == ROUTING_PROTOCOL_STATIC) {
                routing_protocol_name = "static";
            } else if(data[i + 2] == ROUTING_PROTOCOL_OLSR) {
                routing_protocol_name = "OLSR";
            } else {
                if(routing_protocol != 0) {
                    fprintf(stderr, "Unknown routing protocol %d\n",
                            routing_protocol);
                    routing_protocol = 0;
                }
            }
            j = i + 3;
            while(j < i + olen) {
                int oopt = data[j], oolen = data[j + 1];
                if(oolen < 2 || j + oolen > i + olen) {
                    fprintf(stderr, "Truncated suboption.\n");
                    return -2;
                }
                if(oopt == STATIC_DEFAULT_GATEWAY ||
                   oopt == OLSR_MULTICAST_ADDRESS) {
                    char *value;
                    if(oolen % 16 != 2)
                        fprintf(stderr, "Unexpected length for %s\n",
                                routing_protocol == ROUTING_PROTOCOL_OLSR ? 
                                "multicast address" : "default gateway");
                    value = parse_address_list(data + j + 2, oolen - 2);
                    if(routing_protocol == ROUTING_PROTOCOL_OLSR)
                        olsr_multicast_address = value;
                    else
                        static_default_gw = value;
                } else {
                    fprintf(stderr, "Unknown suboption %d\n", oopt);
                    if(oopt & 1) return -2;
                }
                j += oolen;
            }
        } else {
            fprintf(stderr, "Unsupported option %d\n", opt);
            if(opt & 1) return -2;
        }
        i += olen;
    }
    if(start < 0 || config_script[0] == '\0')
        return 1;
    pid = fork();
    if(pid < 0) {
        perror("fork");
        return 0;
    } else if(pid == 0) {
        char buf[50];
        char *action;
        snprintf(buf, 50, "%lu", (unsigned long)getppid());
        setenv("AHCP_DAEMON_PID", buf, 1);
        setenv("AHCP_INTERFACE", interface, 1);
        snprintf(buf, 50, "%d", debuglevel);
        setenv("AHCP_DEBUG_LEVEL", buf, 1);
        setenv("AHCP_NETWORK_NAME", network_name, 1);
        if(routing_protocol_name)
            setenv("AHCP_ROUTING_PROTOCOL", routing_protocol_name, 1);
        if(static_default_gw)
            setenv("AHCP_STATIC_DEFAULT_GATEWAY", static_default_gw, 1);
        if(olsr_multicast_address)
            setenv("AHCP_OLSR_MULTICAST_ADDRESS", olsr_multicast_address, 1);
        if(prefix) {
            setenv("AHCP_PREFIX", prefix, 1);
            free(prefix);
        }
        if(nameserver) {
            if(!nodns)
                setenv("AHCP_NAMESERVER", nameserver, 1);
            free(nameserver);
        }
        switch(start) {
        case 0: action = "stop"; break;
        case 1: action = "start"; break;
        default: abort();
        }
        fprintf(stderr, "%s\n", config_script);
        execl(config_script, config_script, action, NULL);
        perror("exec failed");
        exit(42);
    } else {
        int status;
        pid = waitpid(pid, &status, 0);
        if(pid < 0)
            perror("wait");
        if(!WIFEXITED(status))
            fprintf(stderr, "Child died violently (%d)\n", status);
        if(WEXITSTATUS(status) != 0)
            fprintf(stderr, "Child returned error status %d\n",
                    WEXITSTATUS(status));
    }
    return 1;
}

int
configure(const unsigned char *data, int dlen, int save)
{
    char buf[200];
    int fd, rc;

    rc = parse_message(data, dlen, 1);
    if(rc < 0)
        return -1;

    if(save && persistent) {
        snprintf(buf, 200, "%s%d", data_file, getpid());
        fd = open(buf, O_WRONLY | O_EXCL, 0600);
        if(fd < 0) {
            perror("open(temp)");
            return -1;
        }

        rc = write(fd, data, dlen);
        if(rc < dlen) {
            if(rc < 0)
                perror("write(temp)");
            else
                fprintf(stderr, "write(temp): short write\n");
            close(fd);
            unlink(buf);
            return -1;
        }

        close(fd);

        rc = rename(buf, data_file);
        if(rc < 0) {
            perror("rename(temp, data)");
            unlink(buf);
            return -1;
        }
    }
    return 1;
}

int
unconfigure(const unsigned char *data, int len)
{
    int rc;
    rc = parse_message(data, len, 0);
    return rc;
}

static void 
exitsignal(int signum)
{
    exiting = 1;
}

static void
confsignal(int signum)
{
    reconfigure = 1;
}

int
read_data(unsigned char **data_return, int *dlen_return,
          unsigned long long *serial_return, unsigned int *ttl_return)
{
    int fd, rc;
    struct stat ss;
    unsigned char *data;
    int dlen;

    fd = open(data_file, O_RDONLY);
    if(fd < 0) {
        if(errno == ENOENT) {
            fprintf(stderr, "No data available.\n");
            return 0;
        } else {
            perror("open(data)");
            return -1;
        }
    } else {
        rc = fstat(fd, &ss);
        if(rc < 0) {
            perror("stat(data)");
            return -1;
        }
        if(ss.st_size > 2048) {
            fprintf(stderr, "Data file too long\n");
            return -1;
        }
        data = malloc(ss.st_size);
        dlen = ss.st_size;
        rc = read(fd, data, ss.st_size);
        if(rc < ss.st_size) {
            perror("read(data)");
            return -1;
        }
        close(fd);
        rc = id_message(data, dlen, serial_return, ttl_return);
        if(rc < 0) {
            free(data);
        } else {
            if(*data_return)
                free(*data_return);
            *data_return = data;
            *dlen_return = dlen;
        }
        return rc;
    }
}

int
main(int argc, char **argv)
{
    char *multicast = "ff04::cca6:c0f9:e182:5358";
    int authority = 0;
    int port = 5358;
    unsigned int ttl = 0;
    unsigned long long serial = 0;
    unsigned char *data = NULL;
    int dlen = 0;
    unsigned char zero_message[13] = {PROTOCOL_VERSION,
                                      0, 0, 0, 0, 0, 0, 0, 0,
                                      0, 0, 0, 0};
    unsigned int ifi;
    int configured = 0;
    int rc, s, fd, state, to;
    int on = 1, off = 0, one = 1;
    struct sockaddr_in6 any, mcast;
    struct ipv6_mreq mreq;
    struct timeval now = {0, 0};
    unsigned int seed;
    unsigned char buf[2048];
    int bogus = 0;

    char *syntax = 
        "Syntax: ahcpd "
        "[-aN] [-i interface] [-c script] [-d debuglevel]\n"
        "              [-n name] [-m address] [-p port]\n"
        "              interface\n";

    while(1) {
        int c;
        c = getopt(argc, argv, "ac:d:D:i:m:n:Np:");
        if(c < 0) break;
        switch(c) {
        case 'a': authority = 1; break;
        case 'c': config_script = optarg; break;
        case 'd': debuglevel = atoi(optarg); break;
        case 'D': data_file = optarg; break;
        case 'i': interface = optarg; break;
        case 'm': multicast = optarg; break;
        case 'n': network_name = optarg; break;
        case 'N': nodns = 1; break;
        case 'p': port = atoi(optarg); break;
        default: die(syntax); break;
        }
    }

    if(argc != optind + 1)
        die(syntax);

    interface = argv[optind++];

    fd = open("/dev/urandom", O_RDONLY);
    if(fd < 0) {
        perror("open(random)");
        gettimeofday(&now, NULL);
        seed = now.tv_sec ^ now.tv_usec;
        srandom(seed);
    } else {
        rc = read(fd, &seed, sizeof(unsigned int));
        if(rc < sizeof(unsigned int)) edie("read(random)");
        rc = read(fd, buf, 128);
        if(rc < 128) edie("read(random)");
        close(fd);
        initstate(seed, (char*)buf, 128);
    }

    rc = read_data(&data, &dlen, &serial, &ttl);
    if(rc < 0)
        exit(1);

    ifi = if_nametoindex(interface);
    if(ifi == 0) die("Unknown interface\n");

    s = socket(AF_INET6, SOCK_DGRAM, 0);
    if(s < 0) edie("socket");

    rc = setsockopt(s, IPPROTO_IPV6, IPV6_V6ONLY,
                    (char *)&on, sizeof(on));
    if(rc < 0) edie("setsockopt(IPV6_V6ONLY)");

    rc = setsockopt(s, IPPROTO_IPV6, IPV6_MULTICAST_LOOP,
                    (char*)&off, sizeof(off));
    if(rc < 0) edie("setsockopt(IPV6_MULTICAST_LOOP)");

    rc = setsockopt(s, IPPROTO_IPV6, IPV6_MULTICAST_HOPS,
                    (char*)&one, sizeof(one));
    if(rc < 0) perror("setsockopt(IPV6_MULTICAST_HOPS)");

    rc = setsockopt(s, IPPROTO_IPV6, IPV6_MULTICAST_IF,
                    (char*)&ifi, sizeof(ifi));
    if(rc < 0) perror("setsockopt(IPV6_MULTICAST_IF)");

    memset(&mcast, 0, sizeof(mcast));
    mcast.sin6_family = AF_INET6;
    mcast.sin6_port = htons(port);

    inet_pton(AF_INET6, multicast, mcast.sin6_addr.s6_addr);

    mreq.ipv6mr_interface = ifi;
    mreq.ipv6mr_multiaddr = mcast.sin6_addr;

    memset(&any, 0, sizeof(any));
    any.sin6_family = AF_INET6;
    any.sin6_port = htons(port);
    rc = bind(s, (struct sockaddr*)&any, sizeof(any));
    if(rc < 0) edie("bind");

    rc = setsockopt(s, IPPROTO_IPV6, IPV6_JOIN_GROUP,
                    &mreq, sizeof(mreq));
    if(rc < 0) edie("setsockopt(IPV6_JOIN_GROUP)");

    signal(SIGTERM, exitsignal);
    signal(SIGINT, exitsignal);
    signal(SIGHUP, exitsignal);
    signal(SIGUSR1, confsignal);

    if(authority) {
        rc = configure(data, dlen, 0);
        if(rc < 0)
            die("Couldn't configure authoritative daemon.\n");
        configured = 2;
        fprintf(stderr, "Broadcasting...\n");
        state = STATE_REPLY;
    } else {
        fprintf(stderr, "Querying...\n");
        state = STATE_REQUESTING;
    }
    to = timeout(state);

    gettimeofday(&now, NULL);
    while(1) {
        if(to < 0)
            to = 0;
        rc = waitfd(s, to);
        gettimeofday(&now, NULL);

        if(now.tv_sec < 1165194617 && !bogus) {
            fprintf(stderr, "Your clock is bogus!\n");
            bogus = 1;
        }

        if(exiting) {
            if(configured)
                unconfigure(data, dlen);
            configured = 0;
            break;
        }

        if(reconfigure) {
            if(configured)
                unconfigure(data, dlen);
            configured = 0;
            rc = read_data(&data, &dlen, &serial, &ttl);
            if(rc > 0 && ttl * 10 >= now.tv_sec)
                state = STATE_REPLY;
            else
                state = STATE_REQUESTING;
            to = timeout(state);
            reconfigure = 0;
            continue;
        }

        if(rc >= 0) {
            unsigned long long rserial;
            unsigned int rttl;
            int len;
            to -= rc;
            rc = recv_message(s, &rserial, &rttl, buf, &len);
            if(rc < 0) {
                perror("recv");
                sleep(1);
                continue;
            }

            if(rserial > serial ||
               (ttl * 10 < now.tv_sec && rttl * 10 > now.tv_sec)) {
                rc = parse_message(buf, len, -1);
                if(rc == -2) {
                    fprintf(stderr, "Received an unparseable message.\n");
                    /* But this should still cause us to unconfigure */
                } else if(rc == -1) {
                    fprintf(stderr, "Message not for us.\n");
                    continue;
                }
                
                if(authority) {
                    fprintf(stderr, "Received a conflicting update!\n");
                    continue;
                } else {
                    fprintf(stderr, "Received an update -- reconfiguring.\n");
                    rc = unconfigure(data, dlen);
                    if(rc < 0)
                        fprintf(stderr, "Couldn't unconfigure\n");
                    if(data) free(data);
                    data = NULL;
                    dlen = 0;
                    serial = 0;
                    ttl = 0;
                    configured = 0;
                    rc = configure(buf, len, 1);
                    if(rc >= 0) {
                        data = malloc(len);
                        memcpy(data, buf, len);
                        dlen = len;
                        serial = rserial;
                        ttl = rttl;
                        configured = 2;
                    }
                }
                if(configured <= 0)
                    state = STATE_REQUESTING;
                else if(ttl * 10 >= now.tv_sec)
                    state = STATE_REPLY;
                else
                    state = STATE_IDLE;
                to = timeout(state);
            } else if(rserial == serial) {
                if(configured <= 1) {
                    fprintf(stderr,
                            "Received a confirmation -- "
                            "configuration unchanged.\n");
                    if(configured <= 0)
                        configure(data, dlen, 0);
                    configured = 2;
                }
                if(state != STATE_REPLY) {
                    state = STATE_IDLE;
                    to = timeout(state);
                }
            } else {
                if(state != STATE_REPLY) {
                    if(authority || ttl * 10 >= now.tv_sec)
                        state = STATE_REPLY;
                    else
                        state = STATE_IDLE;
                    to = timeout(state);
                }
            }
        } else {
            if(state == STATE_REQUESTING) {
                rc = send_message(s, &mcast,
                                  zero_message, sizeof(zero_message));
                if(rc < 0)
                    sleep(1);
                state = STATE_REQUESTED;
                to = timeout(state);
            } else if(serial == 0) {
                state = STATE_REQUESTING;
                to = timeout(state);
            } else if(state == STATE_REQUESTED) {
                if(authority) {
                    fprintf(stderr,
                            "STATE_REQUESTED in authoritative daemon "
                            " (this shouldn't happen)\n");
                    state = STATE_REPLY;
                    to = timeout(state);
                } else {                   
                    if(!configured) {
                        fprintf(stderr, 
                                "Warning: no confirmation received, "
                                "using stale data.\n");
                        rc = configure(data, dlen, 0);
                        if(rc >= 0)
                            configured = 1;
                    }
                    if(ttl * 10 >= now.tv_sec)
                        state = STATE_REPLY;
                    else
                        state = STATE_IDLE;
                    to = timeout(state);
                }
            } else if(state == STATE_REPLY || state == STATE_IDLE) {
                if(authority || (ttl * 10 >= now.tv_sec && !bogus))
                    rc = send_message(s, &mcast, data, dlen);
                state = STATE_IDLE;
                to = timeout(state);
            } else 
                abort();
        }
    }
    return 0;
}
